from . import uniprot
from matplotlib import figure
import math

# Fonction d'appel a partir d'un path d'un fichier text
def from_file_to_collection(path_input:str) :
    with open(path_input) as file:
        content = file.read()
    return Collection(content)

class Collection : 
    
    def __init__(self,contenu_uniprot_proteins=None):
        # Sépare notre str brut en autant de fiche uniprot et appel uniprot.build 
        self.uniprot = []
        if not contenu_uniprot_proteins is None:
            contenu_uniprot_proteins = contenu_uniprot_proteins.strip()
            self.liste_txt=contenu_uniprot_proteins.split("//\n")
            self.uniprot_build()

    def uniprot_build(self):
        # A partir d'une liste de str créer notre liste d'objet uniprot
        for elem in self.liste_txt :
            objet_uniprot = uniprot.Uniprot(elem)
            self.uniprot.append(objet_uniprot)

    def add(self, arg:str|uniprot.Uniprot):
        # Utilisé seul pour ajouter un objet uniprot à notre collection
        # Utilisé par la méthode __add__ pour ajouter sequenciellement les objet dans notre nouvelle collection
        if type(arg) == str : 
            object = uniprot.from_file_to_uniprot(arg)
        else :
            object = arg
        if object in self.uniprot : 
            return ValueError
        else :
            self.uniprot.append(object)

    def delet(self,uniprot_id):
        # Retire l'objet uniport donné en argument selon son id
        # Création d'une liste facilement iterable pour les id des objets présents dans notre collection
        liste_id=[]
        for uniprot_object in self.uniprot:
            liste_id.append(uniprot_object.id)
        for uniprot_object in self.uniprot:
            if uniprot_id not in liste_id:
                return ValueError
            else : 
                if uniprot_object.id == uniprot_id:
                    self.uniprot.remove(uniprot_object)
    
    def sort_by_lenght(self):
        # Trier les objet selon la longeur de leur séquence
        objet_taille={}
        for index,objet in enumerate(self.uniprot):
            objet_taille[objet]=len(self.uniprot[index].sequence)
        print(objet_taille)
        self.uniprot = sorted(self.uniprot, key=lambda obj: objet_taille[obj])# key lambda
        return self.uniprot
    
    def filter_for_hydrophobic(self,min_hydro:int):
        # Return les objet uniprot possédant une hydrophibicité global superieur au seil donné en argument
        for uniprot_objet in self.uniprot[:]: # Pour parcourir une copie pas la liste originelle
            if float(uniprot.Uniprot.average_hydrophobicity(uniprot_objet)) < min_hydro:
                self.uniprot.remove(uniprot_objet)
        return self.uniprot
    
    def __add__(self,collection_2):
        #A partir de deux collection créer une nouvelle collection avec tout les objet uniprot unique
        print(self.uniprot)
        print(collection_2.uniprot)
        liste_id = [] # Liste permettant la verifiaction des objets uniques
        new_collection = Collection()
        # Ajout des objet de la collection initial avec la méthode add dans la nouvelle collection vide
        for objet_uniprot in self.uniprot : 
            if objet_uniprot.id not in liste_id :
                liste_id.append(objet_uniprot.id)
                new_collection.add(objet_uniprot)
            else : pass
        # Ajout des objet de la collection 2 avec la méthode add dans la nouvelle collection en evitant les doublons
        for objet_uniprot in collection_2.uniprot : 
            if objet_uniprot.id not in liste_id :
                liste_id.append(objet_uniprot.id)
                new_collection.add(objet_uniprot)
        print(f'La Collection {new_collection} a été créer et contient les objet : \n {new_collection.uniprot}')
        
    def go_view(self):
        # Retourne la liste unique des GO_term au sein de la collection
        go_terms={}
        for uniprot_object in self.uniprot:
            for go in uniprot_object.go_id:
                if go not in go_terms: # Init
                    go_terms[go] = 1
                else: # Incrémentation
                    go_terms[go] += 1 
        return go_terms
    
    def draw_ABRL(self,uniprot_id:str):
        # Créer une représentation graphique de l'abondance relative des AA d'un ojet uniprot au sein de sa collection
        abondance_AA_proteine= {}
        abondance_moyenne_AA = {}
        for uniprot_object in self.uniprot : 
            # Recuperation de l'abondance dans tout les objet
            for AA in uniprot_object.sequence:
                if AA not in abondance_moyenne_AA:
                    abondance_moyenne_AA[AA] = 1
                else:
                    abondance_moyenne_AA[AA] += 1
            # Recuperation de l'abondance dans l'objet en argument
            if uniprot_object.id == uniprot_id:
                for AA in uniprot_object.sequence:
                    if AA not in abondance_AA_proteine:
                        abondance_AA_proteine[AA] = 1
                    else:
                        abondance_AA_proteine[AA] += 1
        abondance_relative={}
        for clé in abondance_AA_proteine:
            abondance_relative[clé]=(abondance_AA_proteine[clé],abondance_moyenne_AA[clé])
        liste_AA=[]
        liste_abondance_relative=[]
        # Calcul de l'abondance relative
        for AA in abondance_relative:
            liste_AA.append(AA)
            liste_abondance_relative.append(math.log(abondance_relative[AA][0]/(abondance_relative[AA][1]/len(self.uniprot))))
        # Création du graphique
        uniprot_id_plot=figure.Figure()
        ax = uniprot_id_plot.subplots(1, 1)
        ax.bar(liste_AA,liste_abondance_relative)
        ax.set_title(f'Histograme de l\'abondance relative des acides aminées de la protéine {uniprot_id} au sein de sa collection',y=-0.2)
        ax.set_ylabel('Abondance relative dans la collection')
        ax.set_xlabel('Acides aminées')
        uniprot_id_plot.savefig(f"{uniprot_id}.png",bbox_inches="tight")
    
    def print_attributes(self):
        # Méthode aditionelle pour tester nos méthode et notre parsage
        for objet_uniprot in self.uniprot :
            print(f"ID: {objet_uniprot.id}")
            print(f"AC Number: {objet_uniprot.ac_number}")
            print(f"Organisme: {objet_uniprot.organisme}")
            print(f"Gene Name: {objet_uniprot.gene_name}")
            print(f"Sequence: {objet_uniprot.sequence}")
            print(f"GO ID: {objet_uniprot.go_id}")



# Fonction d'appel a partir d'un path d'un fichier text
def from_file_to_uniprot(path_input:str) :
    with open(path_input) as file:
        content = file.read()
    return Uniprot(content)

class Uniprot :
	def __init__(self,contenu_text):#Argument
		#Initialisation des atributs
		self.id = None
		self.ac_number = None
		self.organisme = None
		self.gene_name = None
		self.sequence = None
		self.go_id = None
		self.contenu_text = contenu_text
		# Appel de la methode de parsage
		self.parse(contenu_text)

	def parse(self,contenu_text):
		# Splitter en ligne pour itérer sur des lignes et non des caractères
		lignes = contenu_text.split("\n") 
		
		# Parsage du fichier 
		fasta=str("")
		liste_go=[]
		for ligne in lignes :
			# Définition de l'attribut id de l'objet de la classe uniprot, on cherche alors la ligne dans le fichier uniprot.txt qui commence par le code ID
			if ligne.startswith("ID"):
				self.id=str(ligne.split()[1])
			# Définition de l'attribut ac_number de l'objet de la classe uniprot, on cherche alors la ligne dans le fichier uniprot.txt qui commence par AC,
			# On rajoute la condition "==None" car on souhaite seulement sélectionné le premier numero d'accession, ainsi une fois que celui-ci sera défini, l'attribut
			# ac_number ne sera alors pas modifier et le code passera à la définition de l'organisme
			elif ligne.startswith("AC") and self.ac_number==None:
				self.ac_number = str(ligne.split()[1][:-1])
			# Définition de l'attribut organisme de l'objet de la classe uniprot, on cherche alors la ligne dans le fichier uniprot.txt qui commence par OS
			elif ligne.startswith("OS"):
				self.organisme=str(f'{ligne.split()[1]} {ligne.split()[2]}')
			# Définition de l'attribut gene_name de l'objet de la classe uniprot, on cherche alors la ligne dans le fichier uniprot.txt qui commence par GN,
			# Ici aussi, la présence de plusieurs ligne commencant par GN nous pousse à sélectionner celui-ci sur le première ligne
			elif ligne.startswith("GN") and self.gene_name==None:
				self.gene_name=str(ligne.split()[1][5:])
			# Définition de l'attribut sequence de l'objet de la classe uniprot, on cherche alors la ligne dans le fichier uniprot.txt qui commence par SQ
			elif ligne.startswith("SQ") or ligne.startswith(" "):
				fasta+=str(ligne)
				fasta2=fasta.split(";")
				self.sequence=str(fasta2[3].replace(" ",""))
			# Définition de l'attribut go_id de l'objet de la classe uniprot, on cherche alors la ligne dans le fichier uniprot.txt qui commence par "DR   GO",
			# dans chauque ligne commencant par ce code, on récupère l'identifiant GO qu'on ajoute dans une liste dédié nommé liste go
			elif ligne.startswith("DR   GO"):
				go=ligne.split(";")[1].strip()
				liste_go.append(go)
				self.go_id=liste_go
			# Les différents slices et split sont utilisés pour sélectionné uniquement les informations d'intérêts dans les lignes
	
	def fasta_dump(self):
		# Définition du nom du fichier
		file_name = F'{str(self.ac_number)}.fasta'
		# Création d'un Fstring avec 
		# en header l'ID, l'organisme, le nom de gène puis la séquence
		contenue = F'>{self.id}\t{self.organisme}\t{self.gene_name}\n{self.sequence}'
		with open(f"{file_name}", 'w') as file :
			file.write(contenue)
		print(f"Le fichier {file_name} contenant la sequence fasta de la protéine a été créer.")
	
	def molecular_weight(self):
		# Cette fonction retourne la masse moléculaire de la protéine à l'aide de sa séquence extraite du fichier .txt et stocker dans la variable self.sequence
		# Grâce au tableau présent sur le GitLab, on peut créer le dictionaire mol_weight_per_AA qui prend en clé le code à une lettre des AA et en valeurs leur masse
		# moléculaire
		mol_weight_per_AA = {
    	"A": 89,   # Alanine
    	"R": 174,  # Arginine
    	"N": 132,  # Asparagine
    	"D": 133,  # Aspartic acid
    	"B": 133,  # Asparagine or aspartic acid
    	"C": 121,  # Cysteine
    	"Q": 146,  # Glutamine
    	"E": 147,  # Glutamic acid
    	"Z": 147,  # Glutamine or glutamic acid
    	"G": 75,   # Glycine
    	"H": 155,  # Histidine
    	"I": 131,  # Isoleucine
    	"L": 131,  # Leucine
    	"K": 146,  # Lysine
    	"M": 149,  # Methionine
    	"F": 165,  # Phenylalanine
    	"P": 115,  # Proline
    	"S": 105,  # Serine
    	"T": 119,  # Threonine
    	"W": 204,  # Tryptophan
    	"Y": 181,  # Tyrosine
    	"V": 117   # Valine
		}
		# On initie la variable protein_weight qui sera la valeur finale de la masse moléculaire de la protéine
		protein_weight=0
		# On crée une boucle qui parcours la séquence et qui par AA, récupère l'information de sa masse moléculaire (la valeur dans le dictionnaire) et l'ajoute dans
		# la variable protein_weight. Celle ci augmentant au fur et à mesure des cycle jusqu'a obtenir sa valeur finale à la fin du parcours. 
		for AA in self.sequence:
			protein_weight+=float(mol_weight_per_AA[AA])
		return protein_weight

	def average_hydrophobicity(self):
		# Cette fonction retourne l'hydrophobicité moyenne de la protéine à l'aide de sa séquence extraite du fichier .txt et stocker dans la variable self.sequence
		# Grâce au tableau présent sur le GitLab, on peut créer le dictionaire hydrophobicity_per_AA qui prend en clé le code à une lettre des AA et en valeurs leur 
		# hydrophoby scale. Pour les acides aminées ayant une valeur inderterminé d'hydrophobicité (n/a), leur valeur ont arbitrairement été fixé à 0.
		hydrophobicity_per_AA = {
    	"A": 0.33,    # Alanine
    	"R": 1.00,    # Arginine
    	"N": 0.43,    # Asparagine
    	"D": 2.41,    # Aspartic acid
    	"B": 0,       # Asparagine or aspartic acid (n/a)
    	"C": 0.22,    # Cysteine
    	"Q": 0.19,    # Glutamine
    	"E": 1.61,    # Glutamic acid
    	"Z": 0,       # Glutamine or glutamic acid (n/a)
    	"G": 1.14,    # Glycine
    	"H": -0.06,   # Histidine (deux valeurs possibles, choix de la première)
    	"I": -0.81,   # Isoleucine
    	"L": -0.69,   # Leucine
    	"K": 1.81,    # Lysine
    	"M": -0.44,   # Methionine
    	"F": -0.58,   # Phenylalanine
    	"P": -0.31,   # Proline
    	"S": 0.33,    # Serine
    	"T": 0.11,    # Threonine
    	"W": -0.24,   # Tryptophan
    	"Y": 0.23,    # Tyrosine
    	"V": -0.53    # Valine
		}
		# On initie deux variable total_hydrophobicity et protein_lenght dont la rapport permettra l'obtention de l'hydrophobicité moyenne de la protéine.
		total_hydrophobicity=0
		protein_lenght=float(len(self.sequence))
		for AA in self.sequence:
			total_hydrophobicity+=float(hydrophobicity_per_AA[AA])
		average_hydrophobicity=total_hydrophobicity/protein_lenght
		return average_hydrophobicity
	
	def print_attributes(self) :
	# Méthode aditionelle pour tester nos méthode et notre parsage
		print(f"ID: {self.id}")
		print(f"AC Number: {self.ac_number}")
		print(f"Organisme: {self.organisme}")
		print(f"Gene Name: {self.gene_name}")
		print(f"Sequence: {self.sequence}")
		print(f"GO ID: {self.go_id}")

